-module(eCLock).

-include("eGLock.hrl").
-define(CASE(Cond, Then, That), case Cond of true -> Then; _ -> That end).

-export([
	tryLock/1
	, tryLock/2
	, releaseLock/1
	, lockApply/2
	, lockApply/3
]).

-spec tryLock(KeyOrKeys :: term() | [term()]) -> true | ltimeout.
tryLock(KeyOrKeys) ->
	tryLock(KeyOrKeys, ?LockTimeOut).

tryLock(KeyOrKeys, TimeOut) ->
	case is_list(KeyOrKeys) of
		true ->
			KeyIxs = getKexIxs(KeyOrKeys, []),
			doTryLocks(KeyIxs, TimeOut);
		_ ->
			doTryLock(erlang:phash2(KeyOrKeys, ?eGLockSize), TimeOut)
	end.

doTryLock(KeyIx, TimeOut) ->
	case eNifLock:tryLock(KeyIx) of
		true ->
			true;
		_ ->
			loopLock(KeyIx, TimeOut)
	end.

loopLock(KeyIx, TimeOut) ->
	receive
	after ?ReTryTime ->
		LTimeOut = ?CASE(TimeOut == infinity, TimeOut, TimeOut - ?ReTryTime),
		case LTimeOut >= 0 of
			true ->
				case eNifLock:tryLock(KeyIx) of
					true ->
						true;
					_ ->
						loopLock(KeyIx, LTimeOut)
				end;
			_ ->
				ltimeout
		end
	end.

doTryLocks(KeyIxs, TimeOut) ->
	case tryLockAll(KeyIxs, []) of
		true ->
			true;
		_ ->
			loopLocks(KeyIxs, TimeOut)
	end.

loopLocks(KeyIxs, TimeOut) ->
	receive
	after ?ReTryTime ->
		LTimeOut = ?CASE(TimeOut == infinity, TimeOut, TimeOut - ?ReTryTime),
		case LTimeOut >= 0 of
			true ->
				case tryLockAll(KeyIxs, []) of
					true ->
						true;
					_ ->
						loopLocks(KeyIxs, LTimeOut)
				end;
			_ ->
				ltimeout
		end
	end.

-spec releaseLock(KeyOrKeys :: term() | [term()]) -> ok.
releaseLock(KeyOrKeys) ->
	case is_list(KeyOrKeys) of
		true ->
			KeyIxs = getKexIxs(KeyOrKeys, []),
			[eNifLock:releaseLock(OneKeyIx) || OneKeyIx <- KeyIxs],
			ok;
		_ ->
			eNifLock:releaseLock(erlang:phash2(KeyOrKeys, ?eGLockSize)),
			ok
	end.

-spec lockApply(KeyOrKeys :: term() | [term()], MFAOrFun :: {M :: atom(), F :: atom(), Args :: list()} | {Fun :: function(), Args :: list()}) -> term() | {error, ltimeout} | {error, {lock_apply_error, term()}}.
lockApply(KeyOrKeys, MFAOrFun) ->
	lockApply(KeyOrKeys, MFAOrFun, ?LockTimeOut).

-spec lockApply(KeyOrKeys :: term() | [term()], MFAOrFun :: {M :: atom(), F :: atom(), Args :: list()} | {Fun :: function(), Args :: list()}, TimeOut :: integer() | infinity) -> term().
lockApply(KeyOrKeys, MFAOrFun, TimeOut) ->
	case is_list(KeyOrKeys) of
		true ->
			KeyIxs = getKexIxs(KeyOrKeys, []),
			case doTryLocks(KeyIxs, TimeOut) of
				true ->
					try doApply(MFAOrFun)
					catch C:R:S ->
						{error, {lock_apply_error, {C, R, S}}}
					after
						[eNifLock:releaseLock(OneKeyIx) || OneKeyIx <- KeyIxs],
						ok
					end;
				ltimeout ->
					{error, ltimeout}
			end;
		_ ->
			KeyIx = erlang:phash2(KeyOrKeys, ?eGLockSize),
			case doTryLock(KeyIx, TimeOut) of
				true ->
					try doApply(MFAOrFun)
					catch C:R:S ->
						{error, {lock_apply_error, {C, R, S}}}
					after
						eNifLock:releaseLock(KeyIx),
						ok
					end;
				ltimeout ->
					{error, ltimeout}
			end
	end.

getKexIxs([], IxAcc) -> IxAcc;
getKexIxs([Key | Keys], IxAcc) ->
	KeyIx = erlang:phash2(Key, ?eGLockSize),
	getKexIxs(Keys, ?CASE(lists:member(KeyIx, IxAcc), IxAcc, [KeyIx | IxAcc])).

tryLockAll([], _LockAcc) ->
	true;
tryLockAll([KeyIx | KeyIxs], LockAcc) ->
	case eNifLock:tryLock(KeyIx) of
		true ->
			tryLockAll(KeyIxs, [KeyIx | LockAcc]);
		_ ->
			[eNifLock:releaseLock(OneLock) || OneLock <- LockAcc],
			false
	end.

doApply({M, F, A}) ->
	apply(M, F, A);
doApply({Fun, Args}) ->
	apply(Fun, Args).

	

